# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import datetime
import logging
import os
import time

from cmf.monitor import schema
from cmf.monitor.abstract_monitor import CollectionError
from cmf.monitor.abstract_impala_monitor import AbstractImpalaMonitor
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.impalad.json_parser import JsonParser
from cmf.monitor.impalad.query_monitor import QueryStreamMonitor
from cmf.monitor.json_metrics_utils import add_metrics_impala
from cmf.monitor.role_update import RoleUpdate
from cmf.monitor.stoppable_thread import StoppableThread
from cmf.monitor.impala_query_update import ImpalaQueryUpdate, \
    ImpalaRuntimeProfile
from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics

LOG = logging.getLogger('ImpaladMonitor')

_ROLETYPE = "IMPALAD"

# This is not part of the Impala config -- you just configure the log directory,
# so we just hardcode it here.
_QUERY_LOG_SUBDIR = "profiles"
_QUERY_LOG_FILE_NAME_PREFIX = "impala_profile_log-"

# Section and key names in cloudera-monitor.properties
# Since CM 4.5
_SECTION = 'impalad'
_SERVICE_NAME = 'service_name'
_SERVICE_VERSION = 'service_version'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'
_WEBSERVER_PORT = 'impalad_webserver_port'
# Added in CM 4.6
_EXECUTING_QUERIES_TIMEOUT_SECONDS = 'executing_queries_timeout_seconds'
_QUERY_MONITORING_PERIOD_SECONDS = 'query_monitoring_period_seconds'
# Added in CM 4.6.1
_USE_SSL = 'use_ssl'
_WEBSERVER_USERNAME = 'webserver_username'
_WEBSERVER_PASSWORD = 'webserver_password'

class ImpaladMonitor(AbstractImpalaMonitor):

  def __init__(self,
               daemon,
               registry,
               cgroup_root=None,
               cgroup_name=None,
               query_monitor=None):
    AbstractImpalaMonitor.__init__(self,
                                   daemon,
                                   registry,
                                   cgroup_root,
                                   cgroup_name,
                                   _SECTION,
                                   _WEBSERVER_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors
    self._query_log_dir = None
    self._query_log_file = None
    self._executing_query_ids_url = None
    self._executing_query_profile_url = None
    # This is the timeout we'll use for all communication with the Impala Daemon
    # web server for executing query information. This is ultimately read from
    # the monitor's properties file.
    self._executing_queries_timeout_seconds = 5.0
    # This is the period of the query montioring thread. This is ultimately read
    # from the monitor's properties file.
    self._query_monitoring_period_seconds = 1.0

    # The query_monitor object is just a set of function pointers we use to do
    # Impala query monitoring. It exists to allow for easier testing and for
    # keeping track of some statistics
    if query_monitor is None:
      query_monitor = QueryStreamMonitor()
    self._query_monitor = query_monitor

    # This is a thread from which all query monitring occurs. This thread polls
    # the Impala Daemon web server for executing query information, polls the
    # log file for completed query infomation and then publishes any information
    # it finds to the appropriate firehose.
    self._query_monitoring_thread = None

    # We scrape this from the Impala Daemon metrics page and include it in the
    # messages we send with query information.
    self._impala_version = None

    # Config data
    self._name = None
    self._service_name = None
    self._service_version = None
    self._log_dir = None
    self._use_ssl = False
    self._username = None
    self._password = None

  def __str__(self):
    return "ImpaladMonitor for %s" % (self._name,)

  def ensure_collecting(self):
    AbstractImpalaMonitor.ensure_collecting(self)
    self._start_query_monitoring_thread()

  def stop(self):
    AbstractImpalaMonitor.stop(self)
    self._stop_query_monitoring_thread()

  def _reload_config(self, confparse):
    """Overrides AbstractMonitor. Returns True/False for success"""
    try:
      service = confparse.get(_SECTION, _SERVICE_NAME)
      name = confparse.get(_SECTION, _ROLE_NAME)
      version = confparse.get(_SECTION, _SERVICE_VERSION)
      log_dir = confparse.get(_SECTION, _LOG_DIR)
      host = confparse.get(_SECTION, _HOST)
      web_port = confparse.get(_SECTION, _WEBSERVER_PORT)
      use_ssl = confparse.getboolean_with_default(_SECTION, _USE_SSL, False)
      executing_queries_timeout_seconds = confparse.getfloat_with_default(
        _SECTION, _EXECUTING_QUERIES_TIMEOUT_SECONDS, 5.0)
      query_monitoring_period_seconds = confparse.getfloat_with_default(
        _SECTION, _QUERY_MONITORING_PERIOD_SECONDS, 1.0)
      username = confparse.get_with_default(_SECTION, _WEBSERVER_USERNAME, None)
      password = confparse.get_with_default(_SECTION, _WEBSERVER_PASSWORD, None)
    except ValueError, ex:
      self.record_error(ex)
      return False

    metrics_url = self._compute_metrics_url(host, web_port, use_ssl)
    if metrics_url is None:
      return False
    executing_query_ids_url = \
        self._compute_executing_query_ids_url(host, web_port, use_ssl)
    if executing_query_ids_url is None:
      return False
    executing_query_profile_url = \
        self._compute_executing_query_profile_url(host, web_port, use_ssl)
    if executing_query_profile_url is None:
      return False
    self._set_query_monitoring_thread_period(query_monitoring_period_seconds)

    self._service_name = service
    self._service_version = version
    self._name = name
    self._log_dir = log_dir
    self._username = username
    self._password = password
    self._metrics_url = metrics_url
    self._executing_query_ids_url = executing_query_ids_url
    self._executing_query_profile_url = executing_query_profile_url
    self._executing_queries_timeout_seconds = executing_queries_timeout_seconds
    self._query_monitoring_period_seconds = query_monitoring_period_seconds

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    self._query_log_dir = os.path.join(self._log_dir,
                                       _QUERY_LOG_SUBDIR)
    self._query_log_file = os.path.join(self._query_log_dir,
                                        _QUERY_LOG_FILE_NAME_PREFIX)

    return True

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG, add_mem_metrics=True)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    self.add_unexpected_exit_metrics(update)

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    add_metrics_impala(update, metrics, self._metric_accessors, _ROLETYPE, 'cdh4')

  def parse_metrics(self, json):
    return json_parser.JsonParser(json)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length,
                             stats=self._query_monitor.get_stats())

  def collect(self):
    """Overrides AbstractMonitor. Collect metrics. Invoked asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      self._metrics_tuple = (None, None, None)
      return

    LOG.debug("Fetching Impala Daemon metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url,
                                                        username=self._username,
                                                        password=self._password)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def is_process_to_monitor(self, process):
    if process is None:
      return False
    return process.name == 'impalad'

  def get_subject_type(self):
    return "IMPALAD"

  def _compute_executing_query_ids_url(self, host, port, use_ssl):
    if host is None or port is None:
      return None
    protocol = "http"
    if use_ssl:
      protocol = "https"
    return "%s://%s:%s/inflight_query_ids" % (protocol, host, port)

  def _compute_executing_query_profile_url(self, host, port, use_ssl):
    if host is None or port is None:
      return None
    protocol = "http"
    if use_ssl:
      protocol = "https"
    return "%s://%s:%s/query_profile_encoded" % (protocol, host, port)

  ### Query Monitoring Calls ###

  def _start_query_monitoring_thread(self):
    """Starts the query monitoring thread if necessary."""
    if self._query_monitoring_thread is not None:
      return
    LOG.info("Starting query monitoring thread.")
    self._query_monitoring_thread = StoppableThread(
      "ImpalaDaemonQueryMonitoring",
      self._query_monitoring_period_seconds,
      self._check_for_queries)
    self._query_monitoring_thread.start()

  def _stop_query_monitoring_thread(self):
    """Stops the query monitoring thread if necessary."""
    if self._query_monitoring_thread is None:
      return
    LOG.info("Stopping query monitoring thread.")
    self._query_monitoring_thread.should_stop()
    self._query_monitoring_thread = None

  def _set_query_monitoring_thread_period(self, period):
    query_monitoring_thread = self._query_monitoring_thread
    if query_monitoring_thread is None:
      return
    if period > 0:
      query_monitoring_thread.set_interval(period)
    else:
      # We don't disable the thread for simplicity -- keeping the lifecycle of
      # the thread tied to the lifecycle of the monitor. Instead we adjust the
      # period to just once a minute and make _check_for_queries do nothing.
      query_monitoring_thread.set_interval(60)

  def _get_impala_version(self):
    result, time, duration = self.collect_metrics_from_url(self._metrics_url,
                                                           username=self._username,
                                                           password=self._password)
    if result is None or isinstance(result, CollectionError):
      # We have already logged...
      return None
    return result.json.get('impala-server.version', None)
    
  def _check_for_queries(self):
    """
    This is the main method of the query monitoring thread. It checks for
    executing queries first, gets their profiles, and then checks for completed
    queries. If any information is collected, it is sent to HMON immediately.
    """
    if (self._pid is None or 
        self._executing_query_ids_url is None or
        self._executing_query_profile_url is None or
        self._query_monitoring_period_seconds <= 0):
      return

    if self._impala_version is None:
      self._impala_version = self._get_impala_version()
    if self._impala_version is None:
      return

    executing_query_ids = self._query_monitor.get_executing_query_ids(
      self._executing_query_ids_url,
      self._executing_queries_timeout_seconds,
      self._username,
      self._password)

    executing_query_profile_map = {}
    for query_id in executing_query_ids:
      profile = self._query_monitor.get_executing_query_profile(
        self._executing_query_profile_url,
        query_id,
        self._executing_queries_timeout_seconds,
        self._username,
        self._password)
      if profile is not None:
        executing_query_profile_map[query_id] = ImpalaRuntimeProfile(profile)

    completed_query_ids, completed_query_profiles = \
        self._get_completed_query_profiles()

    # We take a quick pass through the completed profiles to see if any of the
    # completed queries are actually done now. We remove them if they are.
    for query_id in completed_query_ids:
      executing_query_profile_map.pop(query_id, None)

    if (len(executing_query_profile_map) > 0 or
        len(completed_query_profiles) > 0):
      self._daemon.send_impala_query_update(
        ImpalaQueryUpdate(role_name=self._name,
                          service_name=self._service_name,
                          impala_version=self._impala_version,
                          executing_query_profiles=\
                            executing_query_profile_map.values(),
                          completed_query_profiles=\
                            completed_query_profiles))

  ### Completed Query Monitoring Calls ###

  def _get_completed_query_profiles(self):
    """
    Calls into the query monitor to get information on newly completed queries,
    advancing our state tracking what constitutes "newly completed" queries as
    appropriate.

    This function returns two lists in matching order. The first list has the
    query IDs of the queries, the second is the profiles as ImpalaRuntimeProfile
    objects.
    
    This should only be called by the query monitoring thread.
    """
    query_log_dir = self._query_log_dir
    query_log_file = self._query_log_file
    if query_log_dir is None or query_log_file is None:
      LOG.debug("Unable to fetch completed queries. No log file specified.")
      return [], []

    # We check for the presence of the query log directory to detect whether we
    # should even do query monitoring. This directory was introduced in Impala
    # 1.0.1.
    if not self._query_log_dir_exists(query_log_dir):
      return [], []

    try:
      completed_queries = \
          self._query_monitor.get_completed_queries(query_log_file)

      completed_query_ids = []
      completed_query_profiles = []
      for query_id, profile in completed_queries:
        completed_query_ids.append(query_id)
        completed_query_profiles.append(ImpalaRuntimeProfile(profile))

      return completed_query_ids, completed_query_profiles

    except Exception, ex:
      LOG.exception("Error fetching completed queries from '%s'" % 
                    (query_log_file,))
      return [], []

  # To enable easier testing.
  def _query_log_dir_exists(self, query_log_dir):
    return os.path.isdir(query_log_dir)
